#!/usr/bin/env python3
"""Matrix ASGI Server."""
import asyncio
import logging
from signal import SIGINT, SIGTERM
from typing import Any

from nio import AsyncClient
from nio.events.room_events import RoomMessageText
from nio.exceptions import LocalProtocolError
from nio.responses import JoinError, RoomSendError
from nio.rooms import MatrixRoom

from markdown import markdown

from . import conf, utils

LOGGER = logging.getLogger("matrix-asgi.server")


class AsgiMatrixServer:
    """Matrix ASGI Server."""

    def __init__(self, /, application="", verbosity=0):
        """Initialize."""
        self.args = conf.get_parser(__doc__).parse_args()
        if verbosity:
            self.args.verbose = verbosity
        if application:
            self.args.application = application

        logging.basicConfig(level=50 - 10 * self.args.verbose)

        self.client = AsyncClient(self.args.matrix_url, self.args.matrix_id)
        self.client.add_event_callback(self.message_callback, RoomMessageText)
        self.application = utils.get_application(self.args.application)

        self.app_scope = {
            "type": "matrix",
            "asgi": {
                "version": "3.0",
                "spec_version": "1.0",
            },
        }
        self.queue = None
        self.event = None

    def run(self):  # pragma: no cover
        """Start the synchronous ASGI server."""
        LOGGER.info("Starting...")
        asyncio.run(self.main())
        LOGGER.info("Stopped")

    async def login(self):
        """Login or re-login on matrix homeserver."""
        await self.client.login(self.args.matrix_pw)

    async def main(self, event=None):
        """Start the asynchronous ASGI server."""
        self.queue = asyncio.Queue()
        self.event = event or asyncio.Event()

        if event is None:  # pragma: no cover
            for sig in (SIGINT, SIGTERM):
                asyncio.get_running_loop().add_signal_handler(
                    sig, utils.terminate, self.event, sig
                )

        await self.login()
        asyncio.create_task(
            self.application(
                self.app_scope, receive=self.queue.get, send=self.app_send
            ),
        )
        asyncio.create_task(
            self.client.sync_forever(timeout=36_000),
        )
        await self.queue.put(
            {
                "type": "matrix.connect",
            }
        )

        LOGGER.info("Started")

        if event is None:  # pragma: no cover
            await self.serve()

    async def serve(self):
        """Server requests for ever."""
        await self.event.wait()
        await self.queue.put(
            {
                "type": "matrix.disconnect",
            }
        )
        LOGGER.info("Stopping...")
        await self.client.close()

    async def app_send(self, message):
        """ASGI application `send` method: Dispatch incomming Channel messages."""
        LOGGER.debug(f"app_send {message=}")
        assert message["type"] == "matrix.send"
        await self.matrix_room_send(message["room"], message["body"])

    async def message_callback(self, room: MatrixRoom, event: RoomMessageText):
        """Handle an incomming message from matrix-nio client: put it in the Queue."""
        LOGGER.debug("message_callback {room=} {event=}")
        await self.queue.put(
            {
                "type": "matrix.receive",
                "room": room.display_name,
                "user": room.user_name(event.sender),
                "body": event.body,
            }
        )

    async def join_room(self, room_id: str) -> bool:  # pragma: no cover
        """Use matrix-nio client to join a room."""
        LOGGER.debug(f"Join room {room_id=}")

        for _ in range(10):
            try:
                resp = await self.client.join(room_id)
                if not isinstance(resp, JoinError):
                    return True
                if resp.status_code == "M_UNKNOWN_TOKEN":
                    LOGGER.warning("Reconnecting")
                    await self.login()
                elif resp.status_code in ["M_FORBIDDEN", "M_CONSENT_NOT_GIVEN"]:
                    LOGGER.error("room access is forbidden")
                    return False
                elif resp.status_code == "M_UNKNOWN":
                    LOGGER.error(f"join error: {resp.transport_response.status}")
                    return False
            except LocalProtocolError as e:
                LOGGER.error(f"Send error: {e}")
            LOGGER.warning("Trying again")
        LOGGER.error("Homeserver not responding")
        return False

    async def send_room_message(self, room_id: str, content: dict[Any, Any]) -> bool:
        """Use matrix-nio client to send a Matrix message to a room."""
        LOGGER.debug(f"Sending room message in {room_id=}: {content=}")

        for _ in range(10):
            try:
                resp = await self.client.room_send(
                    room_id=room_id, message_type="m.room.message", content=content
                )
                if not isinstance(resp, RoomSendError):
                    return True
                if resp.status_code == "M_UNKNOWN_TOKEN":  # pragma: no cover
                    LOGGER.warning("Reconnecting")
                    await self.login()
                else:  # pragma: no cover
                    LOGGER.error(f"room send error {resp=}")
                    return False
            except LocalProtocolError as e:  # pragma: no cover
                LOGGER.error(f"Send error: {e}")
            LOGGER.warning("Trying again")  # pragma: no cover
        LOGGER.error("Homeserver not responding")  # pragma: no cover
        return False  # pragma: no cover

    async def matrix_room_send(self, room, message):
        """Handle an incoming matrix.send message from Channel: forward it to Matrix."""
        content = {
            "msgtype": "m.text",
            "body": message,
            "format": "org.matrix.custom.html",
            "formatted_body": markdown(message, extensions=["extra"]),
        }
        await self.send_room_message(room, content)



"""Configuration."""
import argparse
import os


def get_parser(doc):
    """Configure the argument parser from environment and command line."""
    parser = argparse.ArgumentParser(description=doc, prog="matrix-asgi")
    parser.add_argument(
        "-u",
        "--matrix-url",
        default=os.environ.get("MATRIX_URL", "https://matrix.org"),
        help="matrix homeserver url. Default: `https://matrix.org`. "
        "Environment variable: `MATRIX_URL`",
    )
    parser.add_argument(
        "-i",
        "--matrix-id",
        help="matrix user-id. Required. Environment variable: `MATRIX_ID`",
        **(
            {"default": os.environ["MATRIX_ID"]}
            if "MATRIX_ID" in os.environ
            else {"required": True}
        ),
    )
    parser.add_argument(
        "-p",
        "--matrix-pw",
        help="matrix password. Required. Environment variable: `MATRIX_PW`",
        **(
            {"default": os.environ["MATRIX_PW"]}
            if "MATRIX_PW" in os.environ
            else {"required": True}
        ),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="increment verbosity level",
    )

    parser.add_argument(
        "application",
        help="The ASGI application instance to use as path.to.module:application",
    )
    return parser



"""Utility functions."""
import asyncio
import importlib


def get_application(application_name):
    """Load a callable from a `module.submodule:Object.callable` string.

    copy-paste from https://github.com/sivulich/mqttasgi/blob/master/mqttasgi/utils.py
    """
    module_path, object_path = application_name.split(":", 1)
    application = importlib.import_module(module_path)
    for bit in object_path.split("."):
        application = getattr(application, bit)

    return application


def terminate(event, signal):  # pragma: no cover
    """Close handling stuff."""
    event.set()
    asyncio.get_running_loop().remove_signal_handler(signal)



"""Starting point for the application.

Can be launched from both `python -m matrix_asgi` and `matrix-asgi`.
"""
from .server import AsgiMatrixServer


def main():
    """Start everything."""
    AsgiMatrixServer().run()


if __name__ == "__main__":
    main()



"""Matrix ASGI Consumers."""
from channels.consumer import AsyncConsumer

import logging

logger = logging.getLogger(__name__)


class MatrixConsumer(AsyncConsumer):
    """MatrixConsumer to be subclassed by users."""

    startswith = ""

    async def connect(self):
        """Connect method to be subclassed by users."""
        pass  # pragma: no cover

    async def disconnect(self):
        """Disonnect method to be subclassed by users."""
        pass  # pragma: no cover

    async def receive(self, matrix_message):
        """Receive method to be subclassed by users."""
        pass  # pragma: no cover

    async def matrix_connect(self, event):
        """Initialize connection."""
        await self.connect()

    async def matrix_disconnect(self, event):
        """Initialize connection."""
        await self.disconnect(0)

    async def matrix_receive(self, event):
        """Handle an incoming message from Matrix: forward it to Channels.

        If `startswith` is defined, this will filter out
        messages not starting with `startswith`.
        """
        if self.startswith:
            if event["body"].startswith(self.startswith):
                event["body"] = event["body"].removeprefix(self.startswith)
            else:
                return
        await self.receive(event)

    async def matrix_send(self, room, body):
        """Handle an incoming message from Channels: forward it to Matrix."""
        await self.send({"type": "matrix.send", "room": room, "body": body})



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['matrix_asgi']

package_data = \
{'': ['*']}

install_requires = \
['Markdown>=3.4.1,<4.0.0',
 'channels>=3.0.5,<4.0.0',
 'matrix-nio>=0.19.0,<0.20.0']

entry_points = \
{'console_scripts': ['matrix-asgi = matrix_asgi.__main__:main']}

setup_kwargs = {
    'name': 'matrix-asgi',
    'version': '1.0.0',
    'description': 'ASGI Server for the Matrix protocol',
    'long_description': '# Matrix ASGI\n\n[![Tests](https://github.com/nim65s/matrix-asgi/actions/workflows/test.yml/badge.svg)](https://github.com/nim65s/matrix-asgi/actions/workflows/test.yml)\n[![Lints](https://github.com/nim65s/matrix-asgi/actions/workflows/lint.yml/badge.svg)](https://github.com/nim65s/matrix-asgi/actions/workflows/lint.yml)\n[![Release](https://github.com/nim65s/matrix-asgi/actions/workflows/release.yml/badge.svg)](https://pypi.org/project/matrix-asgi)\n[![pre-commit.ci status](https://results.pre-commit.ci/badge/github/nim65s/matrix-asgi/main.svg)](https://results.pre-commit.ci/latest/github/nim65s/matrix-asgi/main)\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![codecov](https://codecov.io/gh/nim65s/matrix-asgi/branch/main/graph/badge.svg?token=75XO2X5QW0)](https://codecov.io/gh/nim65s/matrix-asgi)\n[![Maintainability](https://api.codeclimate.com/v1/badges/a0783da8c0461fe95eaf/maintainability)](https://codeclimate.com/github/nim65s/matrix-asgi/maintainability)\n[![PyPI version](https://badge.fury.io/py/matrix-asgi.svg)](https://badge.fury.io/py/matrix-asgi)\n\nwith [matrix-nio](https://github.com/poljar/matrix-nio)\n\n## Install\n\n```\npython3 -m pip install matrix-asgi\n```\n\n## Use it in your app\n\nYou can look at the [models.py](https://github.com/nim65s/matrix-asgi/blob/main/tests/django_app/models.py) and\n[consumers.py](https://github.com/nim65s/matrix-asgi/blob/main/tests/django_app/consumers.py) files in the test\napplication for a simple and quick example.\n\n## Start\n\nCreate a matrix user for the bot, and launch this server with the following arguments and/or environment variables\n(environment variables update defaults, arguments take precedence):\n\n```\nmatrix-asgi\n# OR\npython -m matrix_asgi\n```\n\n```\nusage: matrix-asgi [-h] [-u MATRIX_URL] -i MATRIX_ID -p MATRIX_PW [-v] application\n\nMatrix ASGI Server.\n\npositional arguments:\n  application           The ASGI application instance to use as path.to.module:application\n\noptions:\n  -h, --help            show this help message and exit\n  -u MATRIX_URL, --matrix-url MATRIX_URL\n                        matrix homeserver url. Default: `https://matrix.org`.\n                        Environment variable: `MATRIX_URL`\n  -i MATRIX_ID, --matrix-id MATRIX_ID\n                        matrix user-id. Required.\n                        Environment variable: `MATRIX_ID`\n  -p MATRIX_PW, --matrix-pw MATRIX_PW\n                        matrix password. Required.\n                        Environment variable: `MATRIX_PW`\n  -v, --verbose         increment verbosity level\n```\n\n## Unit tests\n\n```\ndocker compose -f test.yml up --exit-code-from tests --force-recreate --build\n```\n\n## JSON Specification\n\nref. [spec.md](https://github.com/nim65s/matrix-asgi/blob/main/spec.md)\n\n## Changes\n\nref. [CHANGELOG.md](https://github.com/nim65s/matrix-asgi/blob/main/CHANGELOG.md)\n',
    'author': 'Guilhem Saurel',
    'author_email': 'guilhem.saurel@laas.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nim65s/matrix-asgi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
